var proxy = {
    debug: false,
    eventEmitter: null,
    modules: {
        crypto: require('crypto'),
        dnsd: require('dnsd'),
        events: require('events'),
        exec: require('child_process').exec,
        fs: require('fs'),
        http: require('http'),
        https: require('https'),
        net: require('net'),
        tls: require('tls'),
        url: require('url'),
        uuid: require('uuid'),
    },

    requests: [], // List of in-progress requests
    dnsServerEndpoint: false,
    httpServerEndpoint: false,
    httpsKeys: {}, // Stores HTTPS keys for various domains
    httpsOptions: {
        SNICallback: function (domain, callback) {
//            console.log('Checking SNI', domain);
            if (proxy.httpsKeys[domain] && (proxy.httpsKeys[domain] !== 'generating')) {
                console.log('existing key', domain);
                return callback(null, proxy.httpsKeys[domain]);
            }

            // Generate keys on the fly
            proxy.generateHttpsKeys(domain).then(function() {
                return callback(null, proxy.httpsKeys[domain]);
            });
        },
        key: false,
        cert: false,
    },
    serverEndpointHttps: false,

    init: function() {
        process.on('uncaughtException', function (err) {
            console.log('[GLOBAL] Uncaught exception');
            console.log(err);
            console.log(err.stack);
        });

        proxy.eventEmitter = new proxy.modules.events.EventEmitter();

        process.chdir('/root/shared/');

        // Network exchange folder
        if (proxy.modules.fs.existsSync('network') === false) {
            proxy.modules.fs.mkdirSync('network');
        }

        if (proxy.modules.fs.existsSync('/root/scratch/vm/ca/index.txt') === false) {
            proxy.modules.fs.writeFileSync('/root/scratch/vm/ca/index.txt', '');
        } else {
            proxy.modules.fs.truncateSync('/root/scratch/vm/ca/index.txt');
        }

        proxy.modules.fs.writeFileSync('/root/scratch/vm/ca/serial', '01');

        proxy.initDnsServerEndpoint();
        proxy.initHttpServerEndpoint();
        proxy.initStratumServerEndpoint();

        try {
            // Generate default key on the fly
            proxy.generateHttpsKeys('charityengine.com').then(function(err) {
                console.log('done generating keys');
                console.log(err);
                proxy.httpsOptions.key = proxy.modules.fs.readFileSync('/root/scratch/vm/ca/certs/charityengine.com.key.pem', 'utf8');
                proxy.httpsOptions.cert = proxy.modules.fs.readFileSync('/root/scratch/vm/ca/certs/charityengine.com.pem', 'utf8');
                proxy.initHttpsServerEndpoint();
            });
        } catch (e) {
            console.log('Unable to init HTTPS', e);
        }
    },
    initDnsServerEndpoint: function() {
        proxy.dnsServerEndpoint = proxy.modules.dnsd.createServer(function(req, res) {
            //console.log(req);
            //proxy.newRequest(req, res, 'dns');
            console.log('DNS query ('+req.connection.type+')', req.question);
            res.end('172.17.0.1');
        }).listen(53, '0.0.0.0');
        console.log('DNS server running at 0.0.0.0:53');

        proxy.dnsServerEndpoint2 = proxy.modules.dnsd.createServer(function(req, res) {
            console.log('dns 2');
            console.log(req);
            proxy.newRequest(req, res, 'dns');
        }).listen(3645, '0.0.0.0');
        console.log('listening dns');
    },
    initHttpServerEndpoint: function() {
        proxy.httpServerEndpoint = proxy.modules.http.createServer(function (req, res) {
            let id = proxy.newRequest(req, res, 'http');
            setTimeout(proxy.monitorRequest, 1000, id);
        })
            .listen(3646, '0.0.0.0', 1023);
        console.log('listening http');
    },
    initHttpsServerEndpoint: function() {
        proxy.serverEndpointHttps = proxy.modules.https.createServer(proxy.httpsOptions, function (req, res) {
            let id = proxy.newRequest(req, res, 'https');
            setTimeout(proxy.monitorRequest, 1000, id);
        })
            .listen(3647, '0.0.0.0', 1023);
        console.log('listening https', 3647);
    },
    initStratumServerEndpoint: function() {
        proxy.serverEndpointStratum = proxy.modules.net.createServer(function(socket) {
            let id = proxy.modules.uuid.v4();
            let counter = 0;
            proxy.eventEmitter.on('response-'+id, function(response, counter) {
                console.log('Response ' + counter + ' sent to socket');
                // Parse as JSON; not sure why this isn't already the case, though
                response = JSON.parse(response);
                //console.log('DEBUG', response, response.content);
                //console.log('DEBUG', Buffer.from(response.content, 'base64').toString());
                socket.write(Buffer.from(response.content, 'base64'));
                //res.write(Buffer.from(response.content, 'base64'));
            });
            socket.on('data', function(data) {
                // Pass an ID to newRequest() to establish a keepalive session
                proxy.newRequest(data, socket, 'stratum', id);
                counter++;
            });
            socket.on('error', function(err) {
                console.log('socket error:', err);
                socket.destroy();
                // Clean up the keepalive directory
                proxy.modules.fs.rmdir('network/' + id, { recursive: true }, function (err) {
                    if (err) {
                        // If recursive delete fails, do it the old fashioned way
                        console.log('No recursive delete support: network/' + id, err);
                        proxy.modules.fs.readdir('network/' + id, function (rerr, files) {
                            if (!rerr) {
                                files.forEach(function (file) {
                                    proxy.modules.fs.unlink('network/' + id + '/' + file, function () {});
                                });
                            }
                            else {
                                console.log('Unable to delete dir: network/' + id, rerr);
                            }
                        });
                    }
                });
            });
            setTimeout(proxy.monitorKeepalive, 1000, id);
        })
        .listen(3333, '0.0.0.0');
        console.log('listening socket', 3333);
    },
    monitorRequest: function(uuid) {
        //console.log('checking request', uuid);

        let file = uuid + '.res';

        // Check if response is large and streaming
        proxy.modules.fs.access('network/' + file + '.0', proxy.modules.fs.constants.R_OK, function (err) {
            if (!err) {
                proxy.requests[uuid].streaming = true;
                return proxy.monitorRequestStreaming(uuid, 0);
            }

            // First part of the stream not found, check if response is small enough to have been completed at once
            proxy.modules.fs.access('network/' + file, proxy.modules.fs.constants.R_OK, function (err) {
                if (!err) {
                    console.log('monitorRequest() handling .res');
                    // Decrypt the request
                    proxy.modules.fs.readFile('network/' + file, function (err, data) {
                        if (data.length < 33) {
                            // File is incomplete, IV is missing - retry later.
                            return setTimeout(proxy.monitorRequest, 1000, uuid);
                        }

                        let response = proxy.getResponseFromFileData(data);
                        proxy.requests[uuid].resolve(response);

                        // Cleanup network folder
                        proxy.modules.fs.unlink('network/' + uuid + '.res', function () {
                        }); // Empty callback, ignore errors
                    });
                } else {
                    setTimeout(proxy.monitorRequest, 1000, uuid);
                }
            });
        });
    },
    monitorRequestStreaming: function(uuid, part) {
        // Check if response is large and streaming
        proxy.modules.fs.access('network/' + uuid + '.res.' + part, proxy.modules.fs.constants.R_OK, function (err) {
            if (!err) {
                // Decrypt the request
                proxy.modules.fs.readFile('network/' + uuid + '.res.' + part, function (err, data) {
                    if (data.length < 33) {
                        // File is incomplete, IV is missing - retry later.
                        return setTimeout(proxy.monitorRequestStreaming, 1000, uuid, part);
                    }

                    let response = proxy.getResponseFromFileData(data);
                    response = JSON.parse(response);
                    console.log('sending streaming response', uuid, part);
                    if (part === 0) {
                        if (proxy.debug) {
                            console.log(response);
                        }
                        // Headers are only written once
                        proxy.requests[uuid].res.writeHead(response.statusCode, response.headers);
                    }
                    proxy.requests[uuid].res.write(Buffer.from(response.content, 'base64'));

                    setTimeout(proxy.monitorRequestStreaming, 10, uuid, part + 1);

                    // Cleanup network folder
                    proxy.modules.fs.unlink('network/' + uuid + '.res.' + part, function () {
                    }); // Empty callback, ignore errors
                });
            } else {
                // Part file not found - check if result is complete
                proxy.modules.fs.access('network/' + uuid + '.res', proxy.modules.fs.constants.R_OK, function (err) {
                    if (!err) {
                        // result is complete
                        proxy.modules.fs.readFile('network/' + uuid + '.res', function (err, data) {
                            if (data.length < 33) {
                                // File is incomplete, IV is missing - retry later.
                                return setTimeout(proxy.monitorRequestStreaming, 1000, uuid, part);
                            }

                            let response = proxy.getResponseFromFileData(data);
                            proxy.requests[uuid].resolve(response);

                            // Cleanup network folder
                            proxy.modules.fs.unlink('network/' + uuid + '.res', function () {
                            }); // Empty callback, ignore errors
                        });

                        return;
                    }

                    // Part file not found, final file also not found. Retry later
                    setTimeout(proxy.monitorRequestStreaming, 1000, uuid, part);
                });
            }
        });
    },
    monitorKeepalive: function(uuid) {
        proxy.modules.fs.readdir('network/' + uuid, function (err, files) {
            if (!err) {
                files.forEach(function (file) {
                    if (file.substr(-4, 4) == '.res') {
                        let fileID = file.slice(0, -4);
                        console.log('monitorRequest() handling .res (keepalive)');
                        // Decrypt the request
                        proxy.modules.fs.readFile('network/' + uuid + '/' + file, function (err, data) {
                            if (data.length < 33) {
                                // File is incomplete, IV is missing - retry later.
                                //return setTimeout(proxy.monitorKeepalive, 1000, uuid);
                            }
                            else {
                                let response = proxy.getResponseFromFileData(data);
                                // We could resolve the request at this point, but
                                // it can cause problems. There's no guarantee that
                                // this requests and responses map 1:1.
                                // So maybe scrap this level of tracking for
                                // keepalive connections? And just emit an event
                                //proxy.requests[uuid].childRequests[fileID].resolve(response);
                              
                                // Notify the keepalive handler of a response
                                proxy.eventEmitter.emit('response-'+uuid, response, fileID);

                                // Cleanup network folder
                                proxy.modules.fs.unlink('network/' + uuid + '/' + file, function () {
                                }); // Empty callback, ignore errors
                            }
                        });
                    }
                });
                setTimeout(proxy.monitorKeepalive, 1000, uuid);
            }
            else {
                console.error('No keepalive directory found for '+uuid, err);
            }
        });
    },
    getResponseFromFileData(data) {
        let dataString;
        let splitLocation;
        let iv;
        try {
            dataString = data.toString();
            splitLocation = dataString.indexOf(':');
            iv = Buffer.from(dataString.substr(0, splitLocation), 'hex');
            let encrypted = data.slice(splitLocation + 1);

            let decipher = proxy.modules.crypto.createDecipheriv('aes-256-ctr', Buffer.from('60ac35dd1573b0be115171f79d6630ba'), iv);
            let decrypted = decipher.update(encrypted);
            decrypted += decipher.final();
            return JSON.parse(decrypted);
        } catch (e) {
            console.log(err);
            console.log(uuid);
            console.log(dataString);
            console.log(iv);
            console.log(e);
            return false;
        }
    },
    finalizeRequest: function (id, req, res, request, keepalive) {
        console.log('[SERVER] New request ' + id + ' for ' + req.url);
        if (proxy.debug) {
            console.log(request.data);
        }

        proxy.writeRequest(id, res, request, keepalive)
        .then(
            function(response) {
                response = JSON.parse(response);
                console.log('request is done', id);
                if (proxy.requests[id].streaming === false && req.headers) {
                    // If streaming, headers are already sent
                    res.writeHead(response.statusCode, response.headers);
                }
                if (keepalive) {
                    // If using keepalive, write to the socket but do not close
                    res.write(Buffer.from(response.content, 'base64'));
                }
                else {
                    res.end(Buffer.from(response.content, 'base64'));
                }
                //console.log('response', response);
                //console.log('response content', response.content);
            },
            function(err) {
                console.log('request failed', err);
                res.end();
            }
        );
    },
    newRequest: function(req, res, type, id) {
        if (proxy.debug) {
            console.log('new connection', type);
        }
        if (Buffer.isBuffer(req)) {
            req = req.toString();
        }

        let keepalive = (id ? true : false);

        let parsedUrl = (req.url ? proxy.modules.url.parse(req.url) : {});
        let headers = (req.headers ? req.headers : {});

        let request = {
            req: req,
            res: res,
            data: { // Must be compatible with node.js http.request options object
                metadata: {
                    type: type,
                },
                object: {
                    headers: headers,
                    hostname: (parsedUrl.hostname ? parsedUrl.hostname : headers.host),
                    method: req.method,
                },
            },
            resolve: null,
            reject: null,
        };

        if (typeof req.path !== 'undefined') {
            request.data.object.path = req.path;
        } else if (typeof req.url !== 'undefined') {
            if (req.url.substr(0, 1) !== '/') {
                console.log('using url', req.url);
                process.exit();
            }
            request.data.object.path = req.url;
        }

        switch (type) {
            case 'https':
                request.data.object.hostname = req.headers.host;
                break;
            case 'stratum':
                // Just pass the request through
                request.data.object = req;
                break;
        }

        if (!id) {
            id = proxy.modules.uuid.v4();
        }

        switch (req.method) {
            case 'POST':
                const postData = [];
                req.on('data', function (data) {
                    postData.push(data);
                });
                // Limit POST data
//                if (postData.length > 1000000) {
//                    req.connection.destroy();
//                }
                req.on('end', function () {
                    request.data.object.postData = Buffer.concat(postData).toString('binary');
                    if (proxy.debug) {
                        console.log('Post data', postData);
                    }
                    proxy.finalizeRequest(id, req, res, request, keepalive);
                });
                break;
            default:
                proxy.finalizeRequest(id, req, res, request, keepalive);
                break;
        }

        return id;
    },
    writeRequest: function (id, res, request, keepalive) {
        let filename = '';
        let encryptedRequest = proxy.encryptRequestData(request.data);
        let promise = null;
        if (keepalive) {
            if (!proxy.requests[id]) {
                proxy.requests[id] = {
                    keepalive: true,
                    childRequests: {},
                    requestCount: 0
                };
                if (!proxy.modules.fs.existsSync('network/' + id)) {
                    // Need the establish the keepalive directory
                    proxy.modules.fs.mkdirSync('network/' + id, { recursive: true });
                    proxy.modules.fs.writeFile('network/' + id + '/keepalive', Date.now().toString(), function (err) {
                        if (err) {
                            console.log('Error while initializing keepalive:', 'network/' + id + '/keepalive', err);
                        }
                    });
                }
            }
            let counter = ++proxy.requests[id].requestCount;
            filename = 'network/' + id + '/' + counter + '.req';
            proxy.requests[id].childRequests[counter] = request;
            promise = new Promise(function(resolve, reject) {
                proxy.requests[id].childRequests[counter].resolve = resolve;
                proxy.requests[id].childRequests[counter].reject = reject;
                proxy.requests[id].childRequests[counter].res = res;
                proxy.requests[id].childRequests[counter].streaming = false;
            });
        }
        else {
            filename = 'network/' + id + '.req';
            proxy.requests[id] = request;
            promise = new Promise(function(resolve, reject) {
                proxy.requests[id].resolve = resolve;
                proxy.requests[id].reject = reject;
                proxy.requests[id].res = res;
                proxy.requests[id].streaming = false;
            });
        }

        proxy.modules.fs.writeFile(filename, encryptedRequest, function (err) {
            if (err) {
                console.log('Error while writing request file:', filename, err);
            }
        });

        return promise;
    },
    encryptRequestData: function(data) {
        let iv = proxy.modules.crypto.randomBytes(16);
        let cipher = proxy.modules.crypto.createCipheriv('aes-256-ctr', Buffer.from('60ac35dd1573b0be115171f79d6630ba'), iv);
        let encrypted = cipher.update(Buffer.from(JSON.stringify(data), 'binary'));
        let finalBuffer = Buffer.concat([encrypted, cipher.final()]);
        return iv.toString('hex') + ':' + finalBuffer.toString('base64');
    },
    generateHttpsKeys: function(domain) {
        if (proxy.httpsKeys[domain] && (proxy.httpsKeys[domain] === 'generating')) {
            console.log('already generating new key', domain);
            return proxy.loadHttpsKeys(domain);
        }

        console.log('generating new key', domain);
        proxy.httpsKeys[domain] = 'generating'; // to prevent multiple keys from being generated in parallel
        return new Promise(function(resolve, reject) {
            // Check if files already exist
            proxy.modules.fs.stat('/root/scratch/vm/ca/certs/' + domain + '.key.pem', function(err, stats) {
                if (err !== null) {
                    // generate new
                    return proxy.generateNewHttpsKeys(domain, resolve, reject);
                }

                proxy.modules.fs.stat('/root/scratch/vm/ca/certs/' + domain + '.pem', function (err, stats) {
                    if (err !== null) {
                        // generate new
                        return proxy.generateNewHttpsKeys(domain, resolve, reject);
                    }
                    else if ((Date.now() < stats.mtimeMs) || (Date.now() - stats.mtimeMs > 364*24*60*60*1000)) {
                        // Certificate is almost a year old or is from the future, generate new
                        return proxy.generateNewHttpsKeys(domain, resolve, reject);
                    }

                    // already exist
                    proxy.handleLoadHttpsKeys(domain, resolve, reject);
                });
            });
        });
    },
    generateNewHttpsKeys: function(domain, resolve, reject) {
        // @todo Needs error checking
        proxy.modules.exec('openssl genrsa -out /root/scratch/vm/ca/certs/' + domain + '.key.pem 2048', function callback(error, stdout, stderr) {
            //console.log(error, stdout, stderr);
            proxy.modules.exec('openssl req -config /root/scratch/vm/ca/openssl.cnf -key /root/scratch/vm/ca/certs/' + domain + '.key.pem -new -sha256 -out /root/scratch/vm/ca/certs/' + domain + '.csr.pem -subj \'/CN=' + domain + '/O=Proxy/C=US/ST=Massachusetts\'', function callback(error, stdout, stderr) {
                //console.log(error, stdout, stderr);
                proxy.modules.exec('openssl ca -batch -config /root/scratch/vm/ca/openssl.cnf -extensions server_cert -days 365 -notext -md sha256 -in /root/scratch/vm/ca/certs/' + domain + '.csr.pem -out /root/scratch/vm/ca/certs/' + domain + '.pem', function callback(error, stdout, stderr) {
                    console.log(error, stdout, stderr);
                    proxy.handleLoadHttpsKeys(domain, resolve, reject);
                });
            });
        });
    },
    loadHttpsKeys: function(domain) {
        return new Promise(function(resolve, reject) {
            proxy.handleLoadHttpsKeys(domain, resolve, reject);
        });
    },
    handleLoadHttpsKeys: function(domain, resolve, reject) {
        if (proxy.httpsKeys[domain] !== 'generating') {
            return resolve(proxy.httpsKeys[domain]);
        }

        proxy.modules.fs.stat('/root/scratch/vm/ca/certs/' + domain + '.key.pem', function(err, stats) {
            if (err !== null) {
                // retry in 500ms
                return setTimeout(proxy.handleLoadHttpsKeys, 500, domain, resolve, reject);
            }

            proxy.modules.fs.stat('/root/scratch/vm/ca/certs/' + domain + '.pem', function(err, stats) {
                if (err !== null) {
                    // retry in 500ms
                    return setTimeout(proxy.handleLoadHttpsKeys, 500, domain, resolve, reject);
                }

                proxy.httpsKeys[domain] = proxy.modules.tls.createSecureContext({
                    key: proxy.modules.fs.readFileSync('/root/scratch/vm/ca/certs/' + domain + '.key.pem', 'utf8'),
                    cert: proxy.modules.fs.readFileSync('/root/scratch/vm/ca/certs/' + domain + '.pem', 'utf8'),
                });
                resolve(proxy.httpsKeys[domain]);
            });
        });
    }
}

module.exports = proxy;

proxy.init();
setTimeout(function() {}, 100000);
